/***************************************************************************
 *   Copyright (C) 2005 by Paul Lutus   *
 *   lutusp@pl-alpha   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "lens.h"

Lens::Lens(RayTracerWidgetCode *p,double cx,double cy,double lensRadius,double leftSphereRadius,double rightSphereRadius,double userThickness,double ior,double lcf,double rcf,double dispersion,bool leftHyp,bool rightHyp,bool symmetrical) {
    parent = p;
    this->cx = cx;
    this->cy = cy;
    this->lensRadius = lensRadius;
    this->leftSphereRadius = leftSphereRadius;
    this->rightSphereRadius = rightSphereRadius;
    this->userThickness = userThickness;
    this->ior = ior;
    this->lcf = lcf;
    this->rcf = rcf;
    this->dispersion = dispersion;
    this->leftHyp = leftHyp;
    this->rightHyp = rightHyp;
    this->symmetrical = symmetrical;
    setup();
}

Lens::Lens(RayTracerWidgetCode *p,string desc) {
    parent = p;
    stringstream ss(desc);
    char c;
    ss >> cx >> c;
    ss >> cy >> c;
    ss >> lensRadius >> c;
    ss >> leftSphereRadius >> c;
    ss >> rightSphereRadius >> c;
    ss >> userThickness >> c;
    ss >> ior >> c;
    ss >> lcf >> c;
    ss >> rcf >> c;
    ss >> dispersion >> c;
    ss >> leftHyp >> c;
    ss >> rightHyp >> c;
    ss >> symmetrical;
    setup();
}

Lens::~Lens() {}

string Lens::describe() {
    stringstream ss;
    ss << cx << ",";
    ss << cy << ",";
    ss << lensRadius << ",";
    ss << leftSphereRadius << ",";
    ss << rightSphereRadius << ",";
    ss << userThickness << ",";
    ss << ior << ",";
    ss << lcf << ",";
    ss << rcf << ",";
    ss << dispersion << ",";
    ss << leftHyp << ",";
    ss << rightHyp << ",";
    ss << symmetrical;
    return ss.str();
}

string Lens::explain() {
    stringstream ss;
    ss << "# X position: " << cx << "\n";
    ss << "# Y position: " << cy << "\n";
    ss << "# Lens Radius: " << lensRadius << "\n";
    ss << "# Left Sphere Radius: " << leftSphereRadius << "\n";
    ss << "# Right Sphere Radius: " << rightSphereRadius << "\n";
    ss << "# Lens Thickness: " << userThickness << "\n";
    ss << "# Index of Refraction: " << ior << "\n";
    ss << "# Left Curvature Factor: " << lcf << "\n";
    ss << "# Right Curvature Factor: " << rcf << "\n";
    ss << "# Dispersion (abbe): " << dispersion << "\n";
    ss << "# Left Hyperboloid: " << leftHyp << "\n";
    ss << "# Right Hyperboloid: " << rightHyp << "\n";
    ss << "# Symmetrical: " << symmetrical << "\n";
    return ss.str();
}

void Lens::setup() {
    // must not aproach zero
    lcf = (lcf < 1e-8)?1e-8:lcf;
    rcf = (rcf < 1e-8)?1e-8:rcf;
    leftOffset = sqrt(leftSphereRadius*leftSphereRadius - lensRadius*lensRadius);
    rightOffset = sqrt(rightSphereRadius*rightSphereRadius - lensRadius*lensRadius);
    //double rsr = rightSphereRadius;
    //rsr = (rsr < 0)?-rsr:rsr;
    //double lsr = leftSphereRadius;
    //lsr = (lsr < 0)?-lsr:lsr;
    userThickness = (userThickness < 0)?0:userThickness;
    internalThickness = 0;
    // pass 1: compute thickness number
    computeBoxSize(-lensRadius,border,internalThickness);
    computeBoxSize(0,border,internalThickness);
    computeBoxSize(lensRadius,border,internalThickness);
    // pass 2: compute borders with thickness applied
    border.left = 0;
    border.right = 0;

    internalThickness += lensRadius * parent->lensMinThickness;
    internalThickness += userThickness;
    computeBoxSize(-lensRadius,border,internalThickness);
    computeBoxSize(0,border,internalThickness);
    computeBoxSize(lensRadius,border,internalThickness);
    border.left -= .001;
    border.right += .001;
    border.top = lensRadius + .001;
    border.bottom = -lensRadius - .001;
}

void Lens::computeBoxSize(double y,Box &border,double &thick) {
    double xa,xb;
    lensXforY(y,0,xa,xb);
    border.left = min(border.left,xa);
    border.left = min(border.left,xb);
    border.right = max(border.right,xa);
    border.right = max(border.right,xb);
    double q = xa-xb;
    if(q < 0) {
        thick = max(thick,-q/2.0);
    }
}

double  Lens::min(double x,double y) {
    return (x<y)?x:y;
}

double  Lens::max(double x,double y) {
    return (x>y)?x:y;
}

void Lens::drawBox(QPainter &pixpainter,bool focused) {
    double x = cx + border.left;
    double y = cy + border.bottom;
    double w = border.right-border.left;
    double h = border.top-border.bottom;
    double dx,dy;
    parent->spaceToDisplay(x,y,dx,dy);
    w = w * parent->dispScale * parent->ySize;
    h = h * parent->dispScale * -parent->ySize;
    QBrush brush((focused)?QColor(parent->lensSelColor):QColor(parent->lensDefColor));
    pixpainter.fillRect ((int)dx,(int)dy,(int)w,(int)h,brush);
}

void Lens::lensXforY(double y,double ccx,double &xa,double &xb) {
    if(leftHyp) {
        xa = RayTraceComputer::hyperbolicLensProfileXforY(y,ccx,lcf,-leftSphereRadius,leftOffset,internalThickness);
    } else {
        xa = RayTraceComputer::sphericalLensProfileXforY(y,ccx,-leftSphereRadius,leftOffset,internalThickness);
    }
    // right
    if(rightHyp) {
        xb = RayTraceComputer::hyperbolicLensProfileXforY(y,ccx,rcf,rightSphereRadius,rightOffset,-internalThickness);
    } else {
        xb = RayTraceComputer::sphericalLensProfileXforY(y,ccx,rightSphereRadius,rightOffset,-internalThickness);
    }
}

void Lens::drawLens(QPainter &pixpainter) {
    int opx0,opy0;
    int opx1,opy1;
    int opx2,opy2;
    double xa,xb;
    int count = 30;
    QColor outline = QColor(parent->lensOutlineColor);
    for(int i = 0; i <= count;i++) {
        double y = lensRadius * ((2.0 * i / (double) count) - 1.0);
        lensXforY(y,cx,xa,xb);
        parent->drawScaledLine(xb,y+cy,opx1,opy1,pixpainter,outline,0,(i != 0));
        parent->drawScaledLine(xa,y+cy,opx2,opy2,pixpainter,outline,0,(i != 0));
        if(i == 0 || i == count) {
            //printf("%.2lf %.2lf %.2lf\n",xa,xb,y+cy);
            //fflush(stdout);
            parent->drawScaledLine(xa,y+cy,opx0,opy0,pixpainter,outline,0,false);
            parent->drawScaledLine(xb,y+cy,opx0,opy0,pixpainter,outline,0,true);
        }
    }
}

bool Lens::inside(double x,double y) {
    return border.inside(x-cx,y-cy);
}

ComplexNum Lens::profile(int prof) {
    double lo = (leftSphereRadius < 0)?-leftOffset:leftOffset;
    double ro = (rightSphereRadius < 0)?-rightOffset:rightOffset;
    double x = ((prof == 0)?cx-lo+internalThickness:cx+ro-internalThickness);
    return ComplexNum(x,cy);
}

double Lens::tangent(int profile,double y) {
    y -= cy;
    double dy;
    double rad,off;
    rad = leftSphereRadius;
    off = leftOffset;
    double cf = lcf;
    if(profile == 1) {
        rad = -rightSphereRadius;
        off = rightOffset;
        cf = rcf;
    }
    if((profile == 0 && leftHyp) || (profile == 1 && rightHyp)) {
        dy= RayTraceComputer::hyperbolicLensProfileDYforY(y,cx,cf,rad,off,-internalThickness);
    } else {
        dy = RayTraceComputer::sphericalLensProfileDYforY(y,cx,rad,off,-internalThickness);
    }
    return dy;
}

void Lens::computeIntersections(int profile,double x1,double y1,double x2,double y2,double ccx,double ccy,double &xa,double &ya,double &xb,double &yb) {
    double srad = (profile == 0)?leftSphereRadius:rightSphereRadius;
    double sr = (srad < 0)?-srad:srad;
    double offset = (profile == 0)?leftOffset:rightOffset;
    double r = 0,off = 0,d = 0,th = 0;
    if((profile == 0 && leftHyp) || (profile == 1 && rightHyp)) {
        r = sqrt(sr*sr-offset*offset);
        off = offset;
        d = sr - off;
        th = internalThickness;
        double cf = rcf;
        if(profile == 0) {
            d = -d;
            th = -th;
            cf = lcf;
        }
        if(srad < 0) {
            d = -d;
        }
        RayTraceComputer::hyp_line_intersections(x1,y1,x2,y2,cf,r,d,cx-th,cy,xa,ya,xb,yb);
    } else {
        RayTraceComputer::circle_line_intersections(x1,y1,x2,y2,sr,ccx,ccy,xa,ya,xb,yb);
    }
}
