/***************************************************************************
 *   Copyright (C) 2005 by Paul Lutus   *
 *   lutusp@pl-alpha   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "raytracecomputer.h"

const double RayTraceComputer::toRadians = M_PI / 180.0;
const double RayTraceComputer::toDegrees = 180.0 / M_PI;
// sodium yellow wavelength nm
const double RayTraceComputer::dispersionPivot = 589.3;
const double RayTraceComputer::dispersionFactor = 500000;

const double RayTraceComputer::redWvlNm = 650;
const double RayTraceComputer::vioWvlNm = 400;

const ColorPos RayTraceComputer::colors[] = {
            ColorPos(1.0,0.0,0.0,650.0), // red
            ColorPos(1.0,1.0,0.0,570.0), // yellow
            ColorPos(0.0,1.0,0.0,510.0), // green
            ColorPos(0.0,0.0,1.0,475.0), // blue
            ColorPos(.5,0.0,1.0,445.0), // indigo
            ColorPos(.5,0.0,.5,400.0), // violet
            ColorPos(.5,0.0,.5,300.0)  // terminator
        };

RayTraceComputer::RayTraceComputer() {}


RayTraceComputer::~RayTraceComputer() {}

vector<string> RayTraceComputer::split(string v,char delim) {
    vector<string> s;
    string::size_type p;
    while((p = v.find(delim)) != string::npos) {
        s.push_back(v.substr(0,p));
        v = v.substr(p+1);
    }
    s.push_back(v);
    return s;
}



ColorPos RayTraceComputer::createColorForWvl(double w) {
    int j;
    for(j = 0;j < 6;j++) {
        if(w > colors[j+1].p && w <= colors[j].p)
            break;
    }
    double r = ntrp(w,colors[j].p,colors[j+1].p,colors[j].r,colors[j+1].r);
    double g = ntrp(w,colors[j].p,colors[j+1].p,colors[j].g,colors[j+1].g);
    double b = ntrp(w,colors[j].p,colors[j+1].p,colors[j].b,colors[j+1].b);
    r *= 255;
    g *= 255;
    b *= 255;
    return ColorPos(r,g,b,0);
}

// dior = ior + ((dp - w) * 500000) / (abbe * dp * w^2)

double RayTraceComputer::dispersionIndex(double ior,double wavelength, double abbe) {
    return ior + ((dispersionPivot - wavelength) * dispersionFactor) / (abbe * dispersionPivot * pow(wavelength,2.0));
}

// line (x1,y1 ... x2,y2) circle (radius, center x center y),
// intersection points xa,ya ... xb,yb

void RayTraceComputer::circle_line_intersections(
    double x1,
    double y1,
    double x2,
    double y2,
    double r,
    double cx,
    double cy,
    double &xa,
    double &ya,
    double &xb,
    double &yb) {
    if (y1 == y2) {
        y2 += 1e-15;
    }
    x1 -= cx;
    x2 -= cx;
    y1 -= cy;
    y2 -= cy;

    xa = (x2*y1*pow(y1 - y2,2) - x2*sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2))) + x1*(-(pow(y1 - y2,2)*y2) + sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2)))))/((pow(x1 - x2,2) + pow(y1 - y2,2))*(y1 - y2));

    xa += cx;

    ya = ((x1 - x2)*(-(x2*y1) + x1*y2) + sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2))))/(pow(x1 - x2,2) + pow(y1 - y2,2));

    ya += cy;

    xb = (-(x1*pow(y1 - y2,2)*y2) - x1*sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2))) + x2*(y1*pow(y1 - y2,2) + sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2)))))/((pow(x1 - x2,2) + pow(y1 - y2,2))*(y1 - y2));

    xb += cx;

    yb = ((x1 - x2)*(-(x2*y1) + x1*y2) - sqrt(pow(y1 - y2,2)*(pow(r,2)*(pow(x1 - x2,2) + pow(y1 - y2,2)) - pow(x2*y1 - x1*y2,2))))/(pow(x1 - x2,2) + pow(y1 - y2,2));

    yb += cy;

}

// line (x1,y1 ... x2,y2) curvature factor (cf),
// hyp radius & depth (r, d), center x and y (cx,cy)
// intersection points xa,ya ... xb,yb

void RayTraceComputer::hyp_line_intersections(
    double x1,
    double y1,
    double x2,
    double y2,
    double cf,
    double r,
    double d,
    double cx,
    double cy,
    double &xa,
    double &ya,
    double &xb,
    double &yb) {

    x1 -= cx;
    x2 -= cx;
    y1 -= cy;
    y2 -= cy;

    xa = (-(d*pow(r,2)*pow(x1 - x2,2)) - 2*pow(d,2)*(y1 - y2)*(x2*y1 - x1*y2) + sqrt(-(pow(d,2)*pow(r,2)*pow(x1 - x2,2)*((-1 + 4*cf)*pow(r,2)*pow(x1 - x2,2) - 4*(pow(d,2)*pow(y1 - y2,2) + d*(y1 - y2)*(x2*y1 - x1*y2) + cf*pow(x2*y1 - x1*y2,2))))))/(2.*(cf*pow(r,2)*pow(x1 - x2,2) - pow(d,2)*pow(y1 - y2,2)));

    xa += cx;

    ya = (d*pow(r,2)*pow(x1 - x2,2)*(-y1 + y2) + 2*cf*pow(r,2)*pow(x1 - x2,2)*(-(x2*y1) + x1*y2) + (y1 - y2)*sqrt(-(pow(d,2)*pow(r,2)*pow(x1 - x2,2)*((-1 + 4*cf)*pow(r,2)*pow(x1 - x2,2) - 4*pow(d,2)*pow(y1 - y2,2) - 4*d*(y1 - y2)*(x2*y1 - x1*y2) - 4*cf*pow(x2*y1 - x1*y2,2)))))/(2.*(x1 - x2)*(cf*pow(r,2)*pow(x1 - x2,2) - pow(d,2)*pow(y1 - y2,2)));

    ya += cy;

    xb = -(d*pow(r,2)*pow(x1 - x2,2) + 2*pow(d,2)*(y1 - y2)*(x2*y1 - x1*y2) + sqrt(-(pow(d,2)*pow(r,2)*pow(x1 - x2,2)*((-1 + 4*cf)*pow(r,2)*pow(x1 - x2,2) - 4*(pow(d,2)*pow(y1 - y2,2) + d*(y1 - y2)*(x2*y1 - x1*y2) + cf*pow(x2*y1 - x1*y2,2))))))/(2.*(cf*pow(r,2)*pow(x1 - x2,2) - pow(d,2)*pow(y1 - y2,2)));

    xb += cx;

    yb = (d*pow(r,2)*pow(x1 - x2,2)*(-y1 + y2) + 2*cf*pow(r,2)*pow(x1 - x2,2)*(-(x2*y1) + x1*y2) + (-y1 + y2)*sqrt(-(pow(d,2)*pow(r,2)*pow(x1 - x2,2)*((-1 + 4*cf)*pow(r,2)*pow(x1 - x2,2) - 4*pow(d,2)*pow(y1 - y2,2) - 4*d*(y1 - y2)*(x2*y1 - x1*y2) - 4*cf*pow(x2*y1 - x1*y2,2)))))/(2.*(x1 - x2)*(cf*pow(r,2)*pow(x1 - x2,2) - pow(d,2)*pow(y1 - y2,2)));

    yb += cy;

}

//      x1 (x4 (-y2 + y3) + x3 (y2 - y4)) + x2 (x4 (y1 - y3) + x3 (-y1 + y4))
// xi = ---------------------------------------------------------------------
//                 -((x3 - x4) (y1 - y2)) + (x1 - x2) (y3 - y4)

double RayTraceComputer::line_line_intersection(
    double ax1,double ay1,double ax2,double ay2,double bx1,double by1,double bx2,double by2
) {
    return (ax1*(ay2*(bx1 - bx2) + bx2*by1 - bx1*by2) + ax2*(ay1*(-bx1 + bx2) - bx2*by1 + bx1*by2))/(-((ay1 - ay2)*(bx1 - bx2)) + (ax1 - ax2)*(by1 - by2));
}

//           x2 (-y1 + y2)   xi (-y1 + y2)
// yi = y2 - ------------- + -------------
//            -x1 + x2        -x1 + x2

double RayTraceComputer::y_intercept(
    double x1, double y1, double x2, double y2, double xi
) {
    return y1 - (x1*(-y1 + y2))/(-x1 + x2) + (xi*(-y1 + y2))/(-x1 + x2);
}



double RayTraceComputer::ntrp(double x,double xa, double xb, double ya, double yb) {
    return ((x-xa)/(xb-xa)) * (yb-ya) + ya;
}

void RayTraceComputer::spaceToDisplay(double x, double y,double &dx,double &dy) {
    dx = ((x-xOffset) * dispScale * ySize) + xCenter;
    dy = yCenter - ((y-yOffset) * dispScale * ySize);
}

void RayTraceComputer::displayToSpaceOffset(double dx, double dy,double &x,double &y) {
    x = ((dx - xCenter) / (dispScale * ySize)) + xOffset;
    y = ((yCenter - dy) / (dispScale * ySize)) + yOffset;
}

void RayTraceComputer::displayToSpace(double dx, double dy,double &x,double &y) {
    x = ((dx - xCenter) / (dispScale * ySize));
    y = ((yCenter - dy) / (dispScale * ySize));
}

void RayTraceComputer::scalePoint(double x, double y,int &ix,int &iy) {
    double px,py;
    spaceToDisplay(x,y,px,py);
    ix = (int) px;
    iy = (int) py;
}

void RayTraceComputer::drawScaledPoint(double x, double y,double radius, QPainter &pixpainter,QColor col) {
    QBrush b1( col );
    pixpainter.setBrush(b1);
    pixpainter.setPen(col);
    int ix,iy;
    scalePoint(x,y,ix,iy);
    int r = (int) radius;
    pixpainter.drawEllipse(ix-r/2,iy-r/2,r,r);
}

/*void RayTraceComputer::drawScaledEllipse(double x, double y,double radius, QPainter &pixpainter) {
 
    QColor blue(0,0,255);
    int ox,oy;
    double radians = M_PI / 180.0;
    for(int i = 0; i <= 360;i+= 5) {
        double px = x + cos(i * radians) * radius;
        double py = y + sin(i * radians) * radius;
        drawScaledLine(px,py,ox,oy,pixpainter,blue,(i != 0));
    }
}*/

void RayTraceComputer::drawScaledLine(double x, double y,int &opx,int &opy, QPainter &pixpainter,QColor col,double width,bool draw) {

    int ipx,ipy;
    scalePoint(x,y,ipx,ipy);

    if(draw) {
        QPen p(col,(int)width);
        pixpainter.setPen(p);
        pixpainter.drawLine(opx,opy,ipx,ipy);
    }
    opx = ipx;
    opy = ipy;
}



double RayTraceComputer::sphericalLensProfileXforY(double y, double cx, double sphereRadius, double offset, double thickness) {
    double v = sqrt(sphereRadius*sphereRadius - y*y);
    double q =  ((sphereRadius < 0)? -offset + v + thickness: offset - v + thickness);
    return q + cx;
}

// 1st derivative

double RayTraceComputer::sphericalLensProfileDYforY(double y, double cx, double sphereRadius, double offset, double thickness) {
    double r = sphereRadius;
    double v = -(y/sqrt(pow(r,2) - pow(y,2)));
    if(sphereRadius < 0)
        v = -v;
    return v;
}

double RayTraceComputer::hyperbolicLensProfileXforY (double y,double cx,double cf, double sphereRadius, double offset,double thickness) {
    double r = sqrt(sphereRadius*sphereRadius-offset*offset);
    double rad = sphereRadius;
    rad = (rad < 0)?-rad:rad;
    double d = rad - offset;
    double v = (-(d*pow(r,2)) + sqrt(pow(d,2)*pow(r,2)*((1 - 4*cf)*pow(r,2) + 4*cf*pow(y,2))))/(2.*cf*pow(r,2));
    if(sphereRadius < 0)
        v = -v;
    return thickness + v + cx;
}

// 1st derivative

double RayTraceComputer::hyperbolicLensProfileDYforY (double y,double cx,double cf, double sphereRadius, double offset,double thickness) {
    double r = sqrt(sphereRadius*sphereRadius-offset*offset);
    double rad = sphereRadius;
    rad = (rad < 0)?-rad:rad;
    double d = rad - offset;
    double v = -(2*pow(d,2)*y)/sqrt(pow(d,2)*pow(r,2)*((1 - 4*cf)*pow(r,2) + 4*cf*pow(y,2)));
    if(sphereRadius < 0)
        v = -v;
    return v;
}

