/***************************************************************************
 *   Copyright (C) 2005 by Paul Lutus   *
 *   lutusp@pl-alpha   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "traceanddraw.h"

TraceAndDraw::TraceAndDraw(RayTracerWidgetCode *p) {
    this->parent = p;
    qpm = NULL;
}


TraceAndDraw::~TraceAndDraw() {
    delete(qpm);
    VLI vli = lensVec.begin();
    while(vli < lensVec.end()) {
        delete(*vli);
        vli++;
    }
}

void TraceAndDraw::updateBuffer(bool raytrace) {
    int newXSize = parent->graphicDisplay->width();
    int newYSize = parent->graphicDisplay->height();
    if(qpm == NULL || parent->xSize != newXSize || parent->ySize != newYSize) {
        delete(qpm);
        parent->xSize = newXSize;
        parent->ySize = newYSize;
        qpm = new QPixmap((int)parent->xSize,(int)parent->ySize);
        parent->xCenter = parent->xSize/2;
        parent->yCenter = parent->ySize/2;
        rayTraceProcess(raytrace);
    }
}

void TraceAndDraw::drawBaseline(QPainter &pixpainter) {
    int opx,opy;
    double x1 = 0;
    double x2 = parent->graphicDisplay->width();
    double y = 0;
    double dx1,dx2,dy;
    parent->displayToSpaceOffset(x1,y,dx1,dy);
    parent->displayToSpaceOffset(x2,y,dx2,dy);
    QColor col(parent->yBaselineColor);
    parent->drawScaledLine(dx1,y,opx,opy,pixpainter,col,0,false);
    parent->drawScaledLine(dx2,y,opx,opy,pixpainter,col,0,true);
}

double TraceAndDraw::gridRound(double v,double modulus,bool roundUp) {
    int sign = (v < 0)?-1:1;
    v = fabs(v);
    if(roundUp)
        v = v + modulus;
    v = v - fmod(v,modulus);
    v *= sign;
    return v;
}

void TraceAndDraw::drawGrid(QPainter &pixpainter) {

    int opx,opy;
    double x1 = 0;
    double x2 = parent->graphicDisplay->width();
    double y1 = 0;
    double y2 = parent->graphicDisplay->height();
    double dx1,dx2,dy1,dy2;
    parent->displayToSpaceOffset(x1,y1,dx1,dy1);
    parent->displayToSpaceOffset(x2,y2,dx2,dy2);
    double fact = 2.0; // increases the density of grid lines
    double e = log10(parent->dispScale * fact) - 100;
    e = e - fmod(e,1.0) + 100;
    double step = pow(10.0,-e);
    double xstart = gridRound(dx1,step) - step;
    double xend = dx2 + step;
    double ystart = gridRound(dy2,step) - step;
    double yend = dy1 + step;
    QColor gridCol(parent->gridColor);
    for(double y = ystart;y <= yend;y += step) {
        for(double x = xstart;x <= xend;x += step) {
            parent->drawScaledLine(x,ystart,opx,opy,pixpainter,gridCol,0,false);
            parent->drawScaledLine(x,yend,opx,opy,pixpainter,gridCol,0,true);
        }
    }
    for(double x = xstart;x <= xend;x += step) {
        for(double y = ystart;y <= yend;y += step) {
            parent->drawScaledLine(xstart,y,opx,opy,pixpainter,gridCol,0,false);
            parent->drawScaledLine(xend,y,opx,opy,pixpainter,gridCol,0,true);
        }
    }

    //parent->displayToSpaceOffset(x1,y1,dx1,dy1);

}

void  TraceAndDraw::rayTraceProcess(bool paint) {
    if(parent->widgetOffsetX == -1) {
        parent->getWidgetOffset(parent->graphicDisplay,parent->widgetOffsetX,parent->widgetOffsetY);
    }

    updateBuffer();
    qpm->fill(QColor(parent->dispColor));
    QPainter pixpainter;
    pixpainter.begin(qpm);
    drawBoxes(pixpainter);
    drawGrid(pixpainter);
    drawBaseline(pixpainter);
    drawLenses(pixpainter);
    // add a grid


    traceRays(pixpainter);
    pixpainter.end();
    if(paint) {
        parent->repaint();
    }
}

void TraceAndDraw::traceRays(QPainter &pixpainter) {
    // generate some rays
    int steps = (int) parent->tracingBeams;
    double min = parent->yStartBeamPos;
    double max = parent->yEndBeamPos;
    double fact = (max-min) / (double)(steps-1);
    for(int i = 0;i < steps;i++) {
        double y = min + (i * fact);
        double xSource = parent->xBeamSourceRefPlane;
        double xTarget = parent->xBeamTargetRefPlane;
        double ba = parent->beamAngle * RayTraceComputer::toRadians;
        double dy = (parent->divergingSource)?0:y;
        if(ba != 0.0) {
            dy = dy - (xTarget - xSource) * tan(ba);
        }
        int dispersionBeams = (int) parent->dispersionBeams;
        if(dispersionBeams > 0) {
            for (int i = 0;i < dispersionBeams;i++) {
                double wvl = (((double)i/(dispersionBeams-1)) * (650.0-400.0))+400.0;
                ColorPos cp = parent->createColorForWvl(wvl);
                QColor beamColor((int)cp.r,(int)cp.g,(int)cp.b);
                traceOneRay(xSource,dy,xTarget,y,pixpainter,lensVec,beamColor,wvl);
            }
        } else { // no dispersion
            QColor beamColor(parent->beamColor);
            traceOneRay(xSource,dy,xTarget,y,pixpainter,lensVec,beamColor,0);
        }
    }
}

// for STL vector sort

class sort_lens_intersections {
public:
    bool operator()(const RayLensIntersection& a, const RayLensIntersection& b) const {
        return a.x < b.x;
    }
};

void TraceAndDraw::traceOneRay(double x1, double y1, double x2, double y2,QPainter &pixpainter,VL lensVec,QColor &beamColor,double wavelength) {
    double xa,ya,xb,yb;
    //double xa2,ya2,xb2,yb2;
    int opx,opy;
    ComplexNum v;

    ComplexNum linea(x1,y1);
    ComplexNum lineb(x2,y2);
    double air_ior = 1.0;
    int maxRays = (int) parent->interactions;
    bool intersecting = true;
    int rays = 0;
    //printf("***start\n");
    while(intersecting && rays++ < maxRays) {

        VC intersections;
        VLI vli;
        for(vli = lensVec.begin();vli < lensVec.end();vli++) {
            for(int prof = 0; prof <= 1;prof++) {
                v = (*vli)->profile(prof);
                (*vli)->computeIntersections(prof,linea.x,linea.y,lineb.x,lineb.y,v.x,v.y,xa,ya,xb,yb);
                if((*vli)->inside(xa,ya)) {
                    intersections.push_back(RayLensIntersection(prof,xa,ya,(*vli)));
                }
                if((*vli)->inside(xb,yb)) {
                    intersections.push_back(RayLensIntersection(prof,xb,yb,(*vli)));
                }
            }
        }
        QColor interColor(parent->intersectionColor);
        int pointRad = (int) parent->intersectionDotRadius;
        parent->drawScaledPoint(linea.x,linea.y,pointRad,pixpainter,interColor);
        sort(intersections.begin(),intersections.end(),sort_lens_intersections());
        // search for the next lens surface to the right (+x)
        VCI vci = intersections.begin();
        intersecting = false;
        while(vci < intersections.end()) {
            if ((*vci).x > linea.x + .001) {
                intersecting = true;
                break;
            }
            vci++;
        }
        if(intersecting) {
            RayLensIntersection inter = *vci;
            parent->drawScaledLine(linea.x,linea.y,opx,opy,pixpainter,beamColor,parent->beamWidth,false);
            parent->drawScaledLine(inter.x,inter.y,opx,opy,pixpainter,beamColor,parent->beamWidth,true);

            // beam incident angle
            double ia = atan2(lineb.y - linea.y,lineb.x - linea.x);

            // surface is tangent to lens intersection point
            // and therefore is the first derivative of that point
            double dy = inter.lens->tangent(inter.profile,inter.y);
            // surface angle
            double sa = atan(dy);

            // Snell's Law calculation block
            double abbe = inter.lens->dispersion;
            double mediaIor = (wavelength == 0 || abbe == 0)?inter.lens->ior:parent->dispersionIndex(inter.lens->ior, wavelength, abbe);
            double n1,n2;
            if(inter.profile == 1) { // if air -> lens
                n1 = air_ior;
                n2 = mediaIor;
            } else {
                n1 = mediaIor;
                n2 = air_ior;
            }

            // 1. add surface angle to incident angle
            double a1 = ia + sa;
            // 2. perform Snell's Law calculation
            double a2 = asin(n1 * sin(a1) / n2);
            // 3. subtract surface angle
            a2 -= sa;

            // create new rectangular vector using result angle
            double nlx = cos(a2);
            double nly = sin(a2);

            // update vectors
            linea.x = inter.x;
            linea.y = inter.y;
            lineb.x = linea.x + nlx;
            lineb.y = linea.y + nly;

        } else { // no intersection, so draw terminating lines at right
            double x = parent->graphicDisplay->width();
            double y = 0;
            double dx,dy;
            parent->displayToSpaceOffset(x,y,dx,dy);
            parent->line_line_intersection(linea.x,linea.y,lineb.x,lineb.y,dx,dy,dx,dy+1);
            dy = parent->y_intercept(linea.x,linea.y,lineb.x,lineb.y,dx);
            parent->drawScaledLine(linea.x,linea.y,opx,opy,pixpainter,beamColor,parent->beamWidth,false);
            parent->drawScaledLine(dx,dy,opx,opy,pixpainter,beamColor,parent->beamWidth,true);
        }
    }
}


void TraceAndDraw::makeDefaultLenses() {
    if(lensVec.size() == 0) {
        double ior = 1.52;
        double disp = 59;
        double cf = 0.03;
        bool leftHyp = false;
        bool rightHyp = false;
        bool symmetrical = true;

        lensVec.push_back(new Lens(parent,0,0,2,6,6,0,ior,cf,cf,disp,leftHyp,rightHyp,symmetrical));
        lensVec.push_back(new Lens(parent,4,0,2,-6,-6,0,ior,cf,cf,disp,leftHyp,rightHyp,symmetrical));
        rayTraceProcess(true);
    }
}

void TraceAndDraw::drawLenses(QPainter &pixpainter) {

    VLI vli = lensVec.begin();
    while(vli < lensVec.end()) {
        (*vli)->drawLens(pixpainter);
        vli++;
    }
}

void TraceAndDraw::drawBoxes(QPainter &pixpainter) {
    VLI vli = lensVec.begin();
    while(vli < lensVec.end()) {
        (*vli)->drawBox(pixpainter,(*vli) == parent->boxTarget);
        vli++;
    }
}

string TraceAndDraw::dispRoundNum(double v) {
    char buf[512];
    double av = fabs(v);
    if(av > 100 || av < .1) {
        sprintf(buf,"%4.3e",v);
    } else {
        sprintf(buf,"%4.3lf",v);
    }
    return string(buf);
}

void TraceAndDraw::updateStatusLabel(int mx, int my,bool inWindow) {
    if(inWindow) {
        double x,y;
        parent->displayToSpaceOffset(mx,my,x,y);
        string sx = dispRoundNum(x);
        string sy = dispRoundNum(y);
        string sz = dispRoundNum(parent->dispScale);
        char buf[512];
        sprintf(buf,"X: %s Y: %s Zoom: %s",sx.c_str(),sy.c_str(),sz.c_str());
        parent->statusEdit->setText(buf);
    } else {
        parent->statusEdit->setText("");
    }
}


